// See the file "COPYING" in the main distribution directory for copyright.
#ifndef __SPCAPESMDIRECTWRITER__H__
#define __SPCAPESMDIRECTWRITER__H__
#include <sys/types.h>
#include <sys/socket.h>
#include <string>
#include <cstdio>
#include <arpa/inet.h>
#include <sys/un.h>
#include <unistd.h>
#include <poll.h>
#include <fcntl.h>

#include "zeek/iosource/Packet.h"
#include "zeek/iosource/IOSource.h"
#include "zeek/iosource/PktDumper.h"
#include "zeek/UID.h"
#include "Writer.h"

namespace SPcap {

//Unix socket where ESM is listening
#define ESM_SOCKET_NAME "/var/run/spcap-esm"
#define ESM_SOCKET_BUFFER_SIZE (70 * 1024 * 1024)

/* Maximum size of the retry queue in elements */
#define MAX_RETRY_QUEUE_SIZE 536870912 /* 512 MB */

/* Throttle connection retries to every Nth Write() */
#define CONNECT_RETRY_THROTTLE 100

/* Maximum number of PacketBuff's to drain from the retry queue on each Write() call */
#define MAX_DRAIN_COUNT 10

typedef enum {
	RUNNING,  // normal mode, connected and writing to ESM
	QUEUEING, // ESM connection broken, queueing packet buffers
	DRAINING  // ESM connection restored, draining the retry queue
} CONNECTION_STATE;

/**
* Handle socket communication with ESM directly (non-threaded).

* ESMDirectWriter implements a retry queue when the connection to ESM is down.
* The initial connection state is QUEUEING, until the connection to ESM is establishd. Once
* established, the connection state is DRAINING and the retry queue is drained. After that the
* writer enters the RUNNING state. If the connection is broken, the writer enters the QUEUEING state
* and begins queueing PacketBuff's, with the same logic used to enter DRAINING, then RUNNING.
*
* The retry queue has a maximum limit on the size of the items in it (MAX_RETRY_QUEUE_SIZE).
* When the limit is reached, PacketBuff's are dropped.
*
* While not connected to ESM, connection attempts are made during calls to Write(). To avoid
* spending too much time trying to (re)connect, these attempts are throttled back (CONNECT_RETRY_THROTTLE).
*
* When draining the retry queue, every call to Write() will attempt to drain multiple elements
* from the retry queue and send them to ESM (MAX_DRAIN_COUNT).
*/
class ESMDirectWriter:public Writer<PacketBuff>
{
public:
	ESMDirectWriter(uint64_t bufferSize): Writer<PacketBuff> { bufferSize },
	                conn(-1), is_open(false), is_failed(false), conn_state(QUEUEING)
		{
		SetTimeOut(default_timeout_sec, default_timeout_usec);
		}
	
	virtual ~ESMDirectWriter()
		{
		Close();
		}

	/**
	* Initialize connection to ESM
	*/
	void Init() override
		{
		(void)Connect();
		}

	/**
	* Write packet buffer to ESM
	* @param pb Packet buffer
	* @return The number of bytes written, -1 on error
	*/
	virtual ssize_t Write(std::unique_ptr<PacketBuff> pb) override;

	/**
	* Stop processing, close connection to ESM.
	*/
	void Stop() override
		{
		Close();
		}

	/**
	* Set the write timeout using seconds and u_sec.
	* @param sec Number of seconds in timeout
	* @param u_sec Number of microseconds in timeout
	*/
	void SetTimeOut(int sec, int u_sec) override;

private:
	/**
	* Write a buffer to ESM. Blocks with timeout.
	* @param buf A pointer to data which has to be transmitted.
	* @param len Length of data.
	* @return -1 on error, else len on successful transmission.
	*/
	ssize_t WriteBuffer(void *buf, size_t len);

	/**
	* Connect to ESM
	* @return false on error.
	*/
	bool Connect();

	/**
	* Close the connection to ESM
	*/
	void Close();


	int conn; //Socket Descriptor
	bool is_open;
	bool is_failed; // permanent failure, errors logged
	const int connection_timeout = 1000; // ms
	//Timeout for blocking writes
	const int default_timeout_sec = 1;
	const int default_timeout_usec = 0;
	struct timeval timeout;
	CONNECTION_STATE conn_state;
	std::queue<std::unique_ptr<PacketBuff> > retry_q;
};

} // namespace SPcap

#endif
