// See the file "COPYING" in the main distribution directory for copyright.
#ifndef __SPCAPESMWRITER__H__
#define __SPCAPESMWRITER__H__
#include <sys/types.h>
#include <string>
#include <cstdio>
#include <arpa/inet.h>
#include <sys/un.h>
#include <unistd.h>

#include "zeek/iosource/Packet.h"
#include "zeek/iosource/IOSource.h"
#include "zeek/iosource/PktDumper.h"
#include "zeek/UID.h"
#include "ThreadedWriter.h"
//Writer for streaming packets to ESM (External Storage Manager)
namespace SPcap {
//Unix socket where ESM is listening to.
#define PACKET_MANAGER_SOCK "/var/run/spcap-esm"
#define SOCK_BUFFER (70 * 1024 * 1024)

class ESMWriter;
//Init message for ESMWriter thread.
class ESMInitMessage : public threading::InputMessage<ESMWriter> {
public:
	ESMInitMessage(ESMWriter* esm_writer):InputMessage<ESMWriter>("ESMInit", esm_writer) {}
	bool Process() override;
};

//Handle socket communication with ESM.
class ESMWriter : public ThreadedWriter
{
public:
	ESMWriter(uint64_t bufferSize): ThreadedWriter { bufferSize },
	          conn(-1), is_open(false),
	          is_connecting(false), write_retries_lw(5), write_retries_hw(2)
		{
		SetTimeOut(default_timeout_sec, default_timeout_usec);
		Start(); //Start the tread
		}

	~ESMWriter()
	{
	if (conn != -1)
		close(conn);
	conn = -1;
	}

	/**
	 * Specifies write attempts before failure
	 *  @param val Number of attempts.
	 */
	void SetWriteRetries(int val)
		{
		write_retries_hw = val;
		write_retries_lw = (val > 1)?val/2:val;
		}

	/**
	 * Initialization invoked from Main thread
	 */
	void Init() override
		{
		SendIn(new ESMInitMessage(this));
		}

	/**
	 * Thread Initialization. Connects to ESM
	 * @return false on failure
	 */
	auto OnInit() -> bool; //Executes on child

	void SetTimeOut(int sec, int u_sec) override;

	virtual ssize_t DoWrite(std::unique_ptr<PacketBuff> pb) override;

private:
	int conn; //Socket Descriptor
	bool is_open;
	bool is_connecting;
	int write_retries_lw; //Low water mark:Pause adding packets to queue
	int write_retries_hw; //High water mark: Close and open socket
	const int connection_retries = 5; //Connection retries
	//Timeout for blocking writes
	const int default_timeout_sec = 5;
	const int default_timeout_usec = 0;
	struct timeval timeout;

	/**
	 * Connect to ESM
	 * @return false on error.
	 */
	bool Connect();

	/**
	 * Write Packets to ESM. Blocks with timeout.
	 * @param buf A pointer to data which has to be transmitted.
	 * @param len Length of data.
	 * @return -1 on error, else len on successful transmission.
	 */
	ssize_t WriteBuffer(void *buf, size_t len);

};

} // namespace SPcap

#endif
