#ifndef __SPCAPTHREADEDWRITER__H__
#define __SPCAPTHREADEDWRITER__H__
#include "zeek/threading/SerialTypes.h"
#include "zeek/threading/MsgThread.h"
#include "Writer.h"
#include "zeek/Reporter.h"

namespace SPcap {

enum class FlowState
	{
	RUN = 1, //Start Sending Packets
	PAUSE //Stop packets until next PKTi_START
	};

/**
* Flow control message to Main thread.
*/
class FlowCtrlMessage : public threading::BasicOutputMessage {
public:
	FlowCtrlMessage(FlowState op) : BasicOutputMessage("FlowCtrlMessage")
		{
		state = op;
		}
	bool Process();
	virtual ~FlowCtrlMessage() {}
private:
	FlowState state;
};

/**
* A packet writer that writes in a child thread; also implements flow control
*/
class ThreadedWriter: public threading::MsgThread, public Writer<PacketBuff>
{
public:
	ThreadedWriter(uint64_t bufferSize):MsgThread(), Writer<PacketBuff>{ bufferSize } {}
	virtual ~ThreadedWriter() {}

	void Stop() override
		{
		SignalStop();
		}

	virtual ssize_t Write(std::unique_ptr<PacketBuff> pb) override;

	/**
	* Derived classes must implement the DoWrite method that
	* actually does the writing in the child thread. This method can
	* also use flow control (pause/resume), as needed. At the very least,
	* once initialized, the child thread needs to call SendResumeMessage();
	* @param pb The packet buffer to write
	* @return The number of bytes written, -1 on error
	*/
	virtual ssize_t DoWrite(std::unique_ptr<PacketBuff> pb) = 0;

protected:
	void SendPauseMessage()
		{
		SendOut(new FlowCtrlMessage(FlowState::PAUSE));
		}

	void SendResumeMessage()
		{   
		SendOut(new FlowCtrlMessage(FlowState::RUN));
		}   

private:
	bool OnHeartbeat(double network_time, double current_time) override 
		{
		return true;
		}

	/** Triggered for execution in the child thread just before shutting threads down.
	*  The child thread should finish its operations.
	*/
		bool OnFinish(double network_time) override 
		{
		return true;
		}
};

/**
* Packet message class sent to child writer thread
*/
class PktMessage : public threading::InputMessage<ThreadedWriter>
{
public:
	PktMessage(ThreadedWriter* w, std::unique_ptr<PacketBuff> pb) : threading::InputMessage<ThreadedWriter>("PktMessage", w)
		{
		pBuff = std::move(pb);
		}

	virtual bool Process()
		{
		Object()->DoWrite(std::move(pBuff));
		return true;
		}

	~PktMessage() {}

private:
	std::unique_ptr<PacketBuff> pBuff;
};

} // namespace SPcap

#endif
